(* this file depends on ../b/b2 and therefore indirectly 
 * on ../b/b1 and ../a/a.v 
 *)

Definition a := 2.

Require b2.

(*** Local Variables: ***)
(*** coq-load-path: ("../a" "../b") ***)
(*** End: ***)
